/*
 *  2ooM: The Master of Orion II Reverse Engineering Project
 *  Simple command-line tool to extract LBX archive files.
 *  Copyright (C) 2006-2008 Nick Bowler
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <fnmatch.h>

#include "tools.h"
#include "lbx.h"


static void printusage(void)
{
	puts("usage: lbxtool [-l|-x] [-v] [-f path] [file ...]");
}

static void printhelp(void)
{
	printusage();
	puts("For now, see the man page for detailed help.");
}

static const char *progname;
#define errmsg(fmt, ...) (\
	fprintf(stderr, "%s: " fmt, progname, __VA_ARGS__)\
)

enum {
	MODE_NONE,
	MODE_LIST,
	MODE_EXTRACT,
};

int filematch(char **argv, const char *name)
{
	int i;

	for (i = 0; argv[i]; i++) {
		switch(fnmatch(argv[i], name, 0)) {
		case 0:
			return 0;
		case FNM_NOMATCH:
			break;
		default:
			errmsg("error matching glob: %s.\n", argv[i]);
			return 1;
		}
	}

	return i ? -1: 0;
}

int list(LBX *lbx, const char *name, int verbose, char **argv) {
	size_t nfiles;
	unsigned int i;

	nfiles = lbx_numfiles(lbx);
	if (verbose) {
		printf("Files in archive: %zu\n", nfiles);
	}

	for (i = 0; i < nfiles; i++) {
		struct lbx_statbuf stat;

		lbx_stat(lbx, i, &stat);

		switch (filematch(argv, stat.name)) {
		case -1: continue;
		case  0: break;
		default: return EXIT_FAILURE;
		}

		printf("%s", stat.name);
		if (verbose) {
			printf(" size=%zu bytes", stat.size);
		}

		putchar('\n');
	}

	return EXIT_SUCCESS;
}

int extract_file(LBXfile *f, const struct lbx_statbuf *stat)
{
	int ret = -1;
	size_t rc;
	FILE *of;

	of = fopen(stat->name, "wb");
	if (!of) {
		errmsg("%s: fopen: %s\n",
			stat->name, strerror(errno));
		return -1;
	}

	while (1) {
		unsigned char buf[1024];

		rc = lbx_file_read(f, buf, sizeof buf);
		if (rc == 0) {
			if (lbx_file_eof(f))
				ret = 0;
			break;
		}

		if (fwrite(buf, rc, 1, of) != 1) {
			errmsg("%s: fwrite: %s\n", stat->name, strerror(errno));
			break;
		}

		if (rc < sizeof buf) {
			if (lbx_file_eof(f))
				ret = 0;
			break;
		}
	}

	if (fclose(of) == EOF) {
		errmsg("%s: fclose: %s\n", stat->name, strerror(errno));
		return -1;
	}

	return ret;
}

int extract(LBX *lbx, const char *name, int verbose, char **argv) {
	size_t nfiles;
	unsigned int i;

	nfiles = lbx_numfiles(lbx);
	if (verbose) {
		printf("Files in archive: %zu\n", nfiles);
	}

	for (i = 0; i < nfiles; i++) {
		struct lbx_statbuf stat;
		LBXfile *file;

		lbx_stat(lbx, i, &stat);

		switch (filematch(argv, stat.name)) {
		case -1: continue;
		case  0: break;
		default: return EXIT_FAILURE;
		}

		file = lbx_file_open(lbx, i);
		if (!file) {
			errmsg("failed to open archive member %s: %s.\n",
				stat.name, lbx_strerror());
			continue;
		}

		if (extract_file(file, &stat) == -1)
			return EXIT_FAILURE;
		lbx_file_close(file);
	}

	return EXIT_SUCCESS;
}

int main(int argc, char **argv)
{
	int mode = MODE_NONE, verbose = 0, opt, rc = EXIT_FAILURE;
	struct lbx_pipe_state state = { .f = stdin };
	const char *name = "stdin";
	LBX *lbx;

	static const char         *sopts   = "lxf:i:vV";
	static const struct option lopts[] = {
		{ "list",    0, NULL, 'l' },
		{ "extract", 0, NULL, 'x' },

		{ "file",    1, NULL, 'f' },
		{ "index",   1, NULL, 'i' },

		{ "verbose", 0, NULL, 'v' },

		{ "version", 0, NULL, 'V' },
		{ "usage",   0, NULL, 'U' },
		{ "help",    0, NULL, 'H' },

		{ 0 }
	};

	progname = "lbxtool"; /* argv[0]; */
	while ((opt = getopt_long(argc, argv, sopts, lopts, NULL)) != -1) {
		switch(opt) {
		case 'l':
			mode = MODE_LIST;
			break;
		case 'x':
			mode = MODE_EXTRACT;
			break;
		case 'f':
			name = strrchr(optarg, '/');
			name = name ? name+1 : optarg;

			if (!freopen(optarg, "rb", state.f)) {
				errmsg("failed to open file %s: %m\n", optarg);
				return EXIT_FAILURE;
			}
			break;
		case 'i':
			/* FIXME: Add index file support. */
			break;
		case 'v':
			verbose = 1;
			break;
		case 'V':
			puts(VERSION_BOILERPLATE("lbxtool"));
			return EXIT_SUCCESS;
		case 'U':
			printusage();
			return EXIT_SUCCESS;
		case 'H':
			printhelp();
			return EXIT_SUCCESS;
		default:
			return EXIT_FAILURE;
		}
	}

	if (fseek(state.f, 0, SEEK_CUR) == 0)
		lbx = lbx_open(state.f, &lbx_default_fops, NULL, name);
	else
		lbx = lbx_open(&state, &lbx_pipe_fops, NULL, name);

	if (!lbx) {
		errmsg("failed to open archive: %s.\n", lbx_strerror());
		return EXIT_FAILURE;
	}

	switch (mode) {
	case MODE_LIST:
		rc = list(lbx, name, verbose, &argv[optind]);
		break;
	case MODE_EXTRACT:
		rc = extract(lbx, name, verbose, &argv[optind]);
		break;
	default:
		fprintf(stderr, "%s: you must specify a mode.\n", progname);
	}

	lbx_close(lbx);
	return rc;
}
